
module Dictionnaire(Dictionnaire,creer,rechercherCle,inserer,tailleReelle,taille,insererIntervalle,insererListe,supprimer,supprimerM,rechercherParCle,mapDico,filtrer,insererM,unionCons,toListVal,recherche,imprimer) where

import Prelude       as P
import Data.Foldable as Fold

-- Arbre Binaire de Recherche par Intervalles
data Dictionnaire cle val =   A (Dictionnaire cle val) cle cle val (Dictionnaire cle val)
                            | F
                            deriving (Eq,Ord,Show,Read)


imprimer F = ""
imprimer (A g l u v d) = imprimer g ++ "["++ show l ++".."++ show u ++ "]=>"++ show v ++ imprimer d



-- Creer un arbre vide
creer :: Dictionnaire cle val
creer = F

-- Rechercher une clef dans les intervalles de l'arbre
rechercherCle :: Ord cle =>
	         Dictionnaire cle val -> cle -> Bool
rechercherCle F _ = False
rechercherCle (A g l u _ d) x
  | x < l = rechercherCle g x
  | x > u = rechercherCle d x
  | otherwise = True


-- Inseerer une clef avec son contexte en modifiant potentielement les intervalles

inserer :: (Ord cle, Enum cle, Eq val) =>
	   Dictionnaire cle val -> cle -> val -> Dictionnaire cle val
inserer a x v = insererIntervalle a x x v

insererM :: (Ord cle, Enum cle, Eq val) =>
	   Dictionnaire cle val -> cle -> val -> Maybe (Dictionnaire cle val)
insererM a x v = insererIntervalleM a x x v


-- Somme des longueurs de tous les intervalles
tailleReelle :: (Enum cle) =>
	        Dictionnaire cle val -> Int
tailleReelle F = 0
tailleReelle (A g l u _ d) = tailleReelle  g + tailleReelle  d + (fromEnum u) - (fromEnum l) +1

-- Nombre de noeuds
taille :: Dictionnaire cle val -> Int
taille F = 0
taille (A g _ _ _ d) = taille  g + taille  d + 1




-- Inseerer tout un interval de clefs pour un context



insererIntervalle :: (Ord cle, Enum cle, Eq val) =>
	             Dictionnaire cle val -> cle -> cle -> val -> Dictionnaire cle val
insererIntervalle a x y v = ifNotingDo (insererIntervalleM a x y v) a

ifNotingDo Nothing a  = a
ifNotingDo (Just b) _ = b


insererIntervalleM :: (Ord cle, Enum cle, Eq val) =>
	             Dictionnaire cle val -> cle -> cle -> val -> Maybe (Dictionnaire cle val)
insererIntervalleM F x y w = Just $ A F x y w F
insererIntervalleM (A g l u v d) x y w
   | y < pred l || (y == pred l && v /= w)
        = fmap (\g' -> A g' l u v d) (insererIntervalleM g x y w)
   | x > succ u || (x == succ u && v/=w)
        = fmap (\d' -> A g l u v d') (insererIntervalleM d x y w)
   | x >= l && y <= u
         = if v==w then Just (A g l u v d) else Nothing
   | x<l && y>u  = do
                     assert (v==w)
                     (g',x') <- absorbeDroite g x w
		     (d',y') <- absorbeGauche d y w
		     return ( A g' x' y' w d' )
   | y > u       = do
                     assert (v==w)
		     (d',y') <- absorbeGauche d y w
		     return (A g l y' w d')
   | x < l       = do
                     assert (v==w)
		     (g',x') <- absorbeDroite g x w
		     return (A g' x' u w d)


absorbeDroite :: (Ord cle, Enum cle, Eq val) =>
	             Dictionnaire cle val -> cle -> val -> Maybe (Dictionnaire cle val , cle)
absorbeDroite F x w = Just (F,x)
absorbeDroite (A g l u v d) x w | x < l      = do
                                                  assert (v==w)
                                                  assert (correctVal d w)
                                                  absorbeDroite g x w
                                | x > succ u = fmap (\(d',x') -> (A g l u v d',x')) (absorbeDroite d x w)
                                | x== succ u && v/=w
				                                    = do
                                                  assert (correctVal d w)
                                                  return (A g l u v d , x)
                                | otherwise  = do
                                                  assert (v==w)
                                                  assert (correctVal d w)
                                                  return (g,l)

absorbeGauche :: (Ord cle, Enum cle, Eq val) =>
	             Dictionnaire cle val -> cle -> val -> Maybe (Dictionnaire cle val , cle)
absorbeGauche F x w = Just (F,x)
absorbeGauche (A g l u v d) x w | x > u      = do
                                                  assert (v==w)
                                                  assert (correctVal g w)
                                                  absorbeGauche d x w
                                | x < pred l = fmap (\(g',x') -> (A g' l u v d,x')) (absorbeGauche g x w)
                                | x== pred l && v/=w
				                                    = do
                                                  assert (correctVal g w)
                                                  return (A g l u v d , x)
                                | otherwise  =  do
                                                  assert (v==w)
                                                  assert (correctVal g w)
                                                  return (d,u)


assert True  = return ()
assert False = Nothing

correctVal F _ = True
correctVal (A g _ _ v d) w = v==w && (correctVal g w) && (correctVal d w)




-- Insertion d'une liste de clefs et contexts
insererListe :: (Ord cle, Enum cle, Eq val) =>
	        Dictionnaire cle val -> [(cle,val)] -> Dictionnaire cle val
insererListe  = Fold.foldl (\a (c,v) -> inserer a c v)

insererListeM :: (Ord cle, Enum cle, Eq val) =>
	        Dictionnaire cle val -> [(cle,val)] -> Maybe (Dictionnaire cle val)
insererListeM  = Fold.foldlM (\a (c,v) -> insererM a c v)

-- Suppression d'une clef
supprimer :: (Ord cle, Enum cle, Eq val) =>
	     Dictionnaire cle val -> cle -> Dictionnaire cle val
supprimer  F _ = F
supprimer (A g l u v d) x | x < l  = A (supprimer g x) l u v d
                          | x > u  = A g l u v (supprimer d x)
                          | x == l && x==u = supprimer_noeud g d
                          | x == l =  A g (succ l) u v d
                          | x == u =  A g l (pred u) v d
                          | otherwise = A (A g l (pred x) v F) (pred x) u v d


supprimer_noeud g F = g
supprimer_noeud g (A gg l u v d) = A (supprimer_noeud g gg) l u v d



-- Suppression d'une clef avec resultat Nothing si clef non presente
supprimerM :: (Ord cle, Enum cle, Eq val) =>
	      Dictionnaire cle val -> cle -> Maybe (Dictionnaire cle val)
supprimerM F _ = Nothing
supprimerM (A g l u v d) x | x < l  = fmap (\g'-> A g' l u v d) (supprimerM g x)
                           | x > u  = fmap (\d'-> A g l u v d') (supprimerM d x)
                           | x == l && x==u = Just (supprimer_noeud g d)
                           | x == l =  Just (A g (succ l) u v d)
                           | x == u =  Just (A g l (pred u) v d)
                           | otherwise = Just (A (A g l (pred x) v F) (succ x) u v d)




-- Rechercher une clef et rendre le contenu trouve, rend Nothing si non trouve
rechercherParCle :: Ord cle =>
	            Dictionnaire cle val -> cle -> Maybe val
rechercherParCle F _ = Nothing
rechercherParCle (A g l u v d) x
  | x < l = rechercherParCle g x
  | x > u = rechercherParCle d x
  | otherwise = Just v

recherche :: (Ord cle, Show cle) =>
	     Dictionnaire cle val -> cle ->  val
recherche F c = error ("cle "++(show c) ++ " non trouvée")
recherche (A g l u v d) x
  | x < l = recherche g x
  | x > u = recherche d x
  | otherwise = v


-- pas demande

foldIntervalle :: (Ord cle, Enum cle) =>
	   (a -> cle -> cle -> b -> a) -> a -> Dictionnaire cle b -> a
foldIntervalle _ acc F = acc
foldIntervalle f acc (A g l u v d) = foldIntervalle f (foldIntervalle f (f acc l u v) g) d


-- modifier tous les contenus d'un dictionnaire
mapDico :: (Ord cle, Enum cle, Eq a, Eq b) =>
	   (a->b) ->  Dictionnaire cle a -> Dictionnaire cle b
mapDico f = foldIntervalle (\a x y v -> insererIntervalle a x y (f v)) creer


-- filtrer les contenus d'un dictionnaire
filtrer :: (Ord cle, Enum cle, Eq val) =>
	   (val->Bool) ->  Dictionnaire cle val -> Dictionnaire cle val
filtrer f = foldIntervalle (\a x y v -> if f v then insererIntervalle a x y v else a) creer




----------------- Phase 3 -----------------------



compresser :: (Ord cle, Enum cle, Eq a) =>
	       Dictionnaire cle a -> Dictionnaire cle a
compresser = foldIntervalle insererIntervalle creer


-- Cette fonction n'est pas safe : le dictionnaire resultant n'est pas compresse
insererIntervalleCons :: (Ord cle, Enum cle) =>
                       Dictionnaire cle [val] -> 
                       cle -> cle   ->   val   ->   Dictionnaire cle [val] 
insererIntervalleCons F x y v = A F x y [v] F
insererIntervalleCons (A g l u v d) x y w
  | y < l
       = A (insererIntervalleCons g x y w) l u v d
  | x > u 
       = A g l u v (insererIntervalleCons d x y w)
  | x > l && y < u
       = A (A g l (pred x) v F) x y (w:v) (A F (succ y) u v d)
  | x == l && y < u
       = A g x y (w:v) (A F (succ y) u v d)
  | x > l && y == u
       = A (A g l (pred x) v F) x u (w:v) d
  | x == l && y == u
       = A g l u (w:v) d
  | x == l && y > u
       = A g l u (w:v) (insererIntervalleCons d (succ u) y w)
  | x < l && y == u
       = A (insererIntervalleCons g x (pred l) w) l u (w:v) d
  | x < l && y > u
       = A (insererIntervalleCons g x (pred l) w)
           l u (w:v)
           (insererIntervalleCons d (succ u) y w)
  | x < l && y < u
       = A (insererIntervalleCons g x (pred l) w)
           l y (w:v)
           (A F (succ y) u v d)
  | x > l && y > u
       = A (A g l (pred x) v F)
           x u (w:v)
           (insererIntervalleCons d (succ u) y w)

-- fait l’union de dictionnaire en recuperant
-- la liste des valeurs en cas de conflit
unionCons :: (Ord cle, Enum cle, Eq val) =>
             [Dictionnaire cle val] -> Dictionnaire cle [val]
unionCons l = compresser (P.foldl unionConsBin creer l)
  where
  unionConsBin d1 F = d1
  unionConsBin d1 (A g l u v d) =
     unionConsBin (unionConsBin (insererIntervalleCons d1 l u v) g) d


toListVal :: (Ord cle, Enum cle) => Dictionnaire cle val -> [val]
toListVal = foldIntervalle (\l _ _ x -> x:l) []